from .http_pipe import HTTPPipe
from .nats_pipe import NATSPipe
from .directions import PipeDirections


class PipePool:
    def __init__(self):
        self._pipes = {}
        self._next_pipe_type = None
        self._next_direction = None
        self._next_handler = None
        self._next_models = None
        self._next_pipe_id = None
        self._next_start_point = None
        self._next_end_point = None

    def _add_pipe(self):
        init_params = {
            'handler': self._next_handler,
            'direction': self._next_direction,
            'models': self._next_models,
        }

        if self._next_pipe_type == 'http':
            if self._next_start_point:
                init_params['socket'] = self._next_start_point
            if self._next_end_point:
                init_params['endpoint'] = self._next_end_point

            cls = HTTPPipe
        elif self._next_pipe_type == 'nats':
            cls = NATSPipe

        pipe = cls(self._next_pipe_id, **init_params)
        self._pipes[pipe.name] = pipe

        self._next_pipe_type = None
        self._next_direction = None
        self._next_handler = None
        self._next_models = None
        self._next_pipe_id = None
        self._next_start_point = None
        self._next_end_point = None

    @property
    def pipes(self): return self._pipes

    async def _handler(self, *args):
        print('handler')

    def pipe(self, *args):
        def decorator(func):
            self._next_pipe_type = args[0]
            self._next_pipe_id = args[1]
            self._next_models = args[2]
            self._add_pipe()

            async def wrapper(*args, **kwargs):
                return await func(*args, **kwargs)
            return wrapper
        return decorator

    def recipient(self, *args):
        def decorator(func):
            self._next_direction = PipeDirections.receiving
            self._next_start_point = args[0]
            self._next_handler = func

            async def wrapper(*args, **kwargs):
                return await func(*args, **kwargs)
            return wrapper
        return decorator

    def sender(self, *args):
        def decorator(func):
            self._next_direction = PipeDirections.sending
            self._next_end_point = args[0]
            self._next_handler = func

            async def wrapper(obj, message, **kwargs):
                _message = message
                func_kwargs = {
                    'message': _message,
                }

                func_result = await func(obj, **func_kwargs)
                if func_result:
                    _message = func_result

                print('Sending message...')
            return wrapper
        return decorator


import json

from aiohttp import web
from message_prototypes import BaseMessage
from message_prototypes.pipe_messages import RequestMessage, ResponseMessage

from .pipe import Pipe
from .directions import PipeDirections


class HTTPPipe(Pipe):
    def __init__(self, name, **kwargs):
        self._http_handler = kwargs.get('handler', None)
        self._socket = kwargs.get('socket', None)
        self._direction = kwargs.get('direction', None)
        self._endpoint = kwargs.get('endpoint', None)
        self._models = kwargs.get('models', tuple([]))

        Pipe.__init__(self, name, **kwargs)

    async def _handler(self, *args):
        request = args[0]

        if request.method not in ('POST', 'OPTIONS') or \
                request.path != '/_pipe':
            return web.Response(
                status=400,
                text='HTTP Method or url path is not allowed!')

        if request.method == 'OPTIONS':
            pipe_info = {
                'allowed_models': [m.__name__ for m in self._models],
            }
            return web.Response(
                status=200,
                body=json.dumps(pipe_info),
                content_type='application/json')

        try:
            data = await request.json()
        except Exception:
            return web.Response(
                status=400,
                text='Bad request body!')

        message_model = BaseMessage.detect_model(data)
        if message_model not in self._models:
            return web.Response(
                status=400, 
                text='Incorrect or invalid model!')

        result = await self._http_handler(
            self._app, message=message_model.unpack(data))
        if result.__class__ == ResponseMessage:
            response = result.pack()
        else:
            response = {}

        return web.Response(
            status=200,
            body=json.dumps(response),
            content_type='application/json')

    def __str__(self):
        result = []

        result.append(f'name={self.name}')

        if self._direction in \
            [PipeDirections.sending, PipeDirections.duplex]:
            result.append(f'endpoint={self._endpoint}')

        if self._direction in \
            [PipeDirections.receiving, PipeDirections.duplex]:
            result.append(f'socket={self._socket}')

        return f'[HTTPPipe {" ".join(result)}]'

    async def _init(self):
        if self._direction == PipeDirections.receiving:
            host, port = self._socket.split(':')
            server = web.Server(self._handler)
            runner = web.ServerRunner(server)

            await runner.setup()

            site = web.TCPSite(runner, host, int(port))

            await site.start()



from pyasyncserver.core.logger import get_logger


logger = get_logger(__name__)


class Pipe:
    def __init__(self, name, **kwargs):
        self._pipe_name = name
        self._app = None

    @property
    def name(self): return self._pipe_name

    async def _init(self):
        pass

    async def init_pipe(self, app):
        self._app = app
        
        await self._init()
        logger.info(f'Init pipe: {self}')


import logging, os

LOGGER_FORMAT = os.environ.get(
    'LOGGER_FORMAT',
    '%(asctime)s - %(levelname)s - [%(name)s => %(filename)s => '
    '%(funcName)s(%(lineno)d)] - %(message)s')


def get_file_handler():
    LOG_PATH = os.environ.get('LOG_PATH', './base.log')

    file_handler = logging.FileHandler(LOG_PATH)
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(logging.Formatter(LOGGER_FORMAT))
    return file_handler

def get_stream_handler():
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(logging.Formatter(LOGGER_FORMAT))
    return stream_handler

def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)

    logger.addHandler(get_file_handler())
    logger.addHandler(get_stream_handler())
    return logger


from .pipe import Pipe


class NATSPipe(Pipe):
    def __init__(self, name, **kwargs):
        self._http_handler = kwargs.get('handler', None)

        Pipe.__init__(self, name, **kwargs)

    def __str__(self):
        return f'[NATSPipe name={self.name}]'


import asyncio
import uvloop

from pyasyncserver.core.logger import get_logger


logger = get_logger(__name__)

class App:
    def __init__(self, pipepool):
        uvloop.install()

        self._loop = asyncio.get_event_loop()

        self._pipepool = pipepool

    def add_pipe(self, pipe):
        self._pipes[pipe.name] = pipe

    def init_app(self):
        pass

    async def _run(self):
        for _, pipe in self._pipepool.pipes.items():
            await pipe.init_pipe(self)
        logger.info('Run')

    def run(self):
        self._loop.run_until_complete(self._run())
        self._loop.run_forever()
        self._loop.close()



from setuptools import setup


setup(
    name='pyasyncserver',
    zip_safe=True,
    version='0.1.5',
    description='Modular asynchronous application service',
    url='https://github.com/agratoth/pyasyncserver',
    maintainer='Anton Berdnikov',
    maintainer_email='agratoth@yandex.ru',
    packages=[
        'pyasyncserver',
        'pyasyncserver.core',
        'pyasyncserver.core.logger',
        'pyasyncserver.apps',
        'pyasyncserver.pipes'
    ],
    package_dir={'pyasyncserver': 'pyasyncserver'},
    install_requires=[
        'uvloop>=0.15.2',
        'aiohttp>=3.7.4',
        'py-message-prototypes>=0.2.1'
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires='>=3.8',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
)
